/************************************************************
Module name: AlertIO.C
Written by: Jim Harkins and Jeffrey Richter
Notices: Copyright (c) 1995 Jeffrey Richter
************************************************************/


#include "..\AdvWin32.H"    /* See Appendix B for details. */
#include <windows.h>
#include <windowsx.h>

#pragma warning(disable: 4001)      /* Single-line comment */

#include <tchar.h>
#include <stdio.h>
#include <stdlib.h>
#include "Resource.H"

/////////////////////////////////////////////////////////////


#define BUFFSIZE (32 * 1024)
#define BUFFNUM  4

#define DSTFILENAME      __TEXT("AlertIO.CPY")

// Status of the copy
// Note: This is an ordered list (i.e., < and > operators are
// used on this type).

typedef enum {
   csCopying,
   csDoneReading,
   csDoneWriting,
   csError,
} COPYSTATUS;


HWND g_hwndLB = NULL;

// Data used by copy
HANDLE g_hFileSrc, g_hFileDst;

// hEvent contains the buffer number
OVERLAPPED g_Overlapped[BUFFNUM];   

// Pointer to file copy buffers.
BYTE g_bBuffers[BUFFSIZE * BUFFNUM];

// Offset in source file where next read begins
DWORD  g_dwNextReadOffset = 0;

// Status of the file copy
COPYSTATUS g_CopyStatus = csCopying;

DWORD g_dwLastError = NO_ERROR;

int g_nReadsInProgress     = 0;
int g_nMaxReadsInProgress  = 0;
int g_nCompletedReads      = 0;
int g_nWritesInProgress    = 0;
int g_nMaxWritesInProgress = 0;
int g_nCompletedWrites     = 0;


VOID WINAPI InputCompletion (DWORD fdwError, 
   DWORD cbTransferred, LPOVERLAPPED lpo);

VOID WINAPI OutputCompletion (DWORD fdwError,
   DWORD cbTransferred, LPOVERLAPPED lpo);


/////////////////////////////////////////////////////////////


// This function constructs a string using the format string
// passed and the variable number of arguments and adds the
// string to the list box identified by the global 
// g_hwndLB variable.
void AddStr (LPCTSTR szFmt, ...) {
   TCHAR szBuf[150];
   int nIndex;
   va_list va_params;

   // Make va_params point to the first argument after szFmt.
   va_start(va_params, szFmt);

   // Build the string to be displayed.
   _vstprintf(szBuf, szFmt, va_params);

   do {
      // Add the string to the end of the list box.
      nIndex = ListBox_AddString(g_hwndLB, szBuf);

      // If the list box is full, delete the first item in it.
      if (nIndex == LB_ERR)
         ListBox_DeleteString(g_hwndLB, 0);

   } while (nIndex == LB_ERR);

   // Select the newly added item.
   ListBox_SetCurSel(g_hwndLB, nIndex);

   // Indicate that we're done referencing
   // the variable arguments.
   va_end(va_params);
}


/////////////////////////////////////////////////////////////


void ReadNext (LPOVERLAPPED lpOverlapped) {
   BOOL  fReadOk;
   DWORD dwLastError;
   int nBufSize = BUFFSIZE / ((int) lpOverlapped->hEvent + 1);

   if (csCopying != g_CopyStatus) {
      // Either an error has occurred or we have read past 
      // the end of the file. In either case,
      // let's not start to read a new block from the file.
      return;
   }      
      
   // The file is still being copied.
   
   // Figure out where to start
   // reading the next portion of the file.
   lpOverlapped->Offset = g_dwNextReadOffset;
   lpOverlapped->OffsetHigh = 0;
      
   // Set the global file offset variable so that it knows
   // where it should begin reading the next time.
   g_dwNextReadOffset += nBufSize;
   

   AddStr(__TEXT("%d: Read, Offset=%d, Len=%d."),
      (int) lpOverlapped->hEvent, lpOverlapped->Offset,
      nBufSize);

   // Initiate the alertable read from the
   // file into the appropriate buffer.
   fReadOk = ReadFileEx(g_hFileSrc,
      g_bBuffers + (int) lpOverlapped->hEvent * BUFFSIZE,
      nBufSize, lpOverlapped, InputCompletion);

   if (fReadOk) {
      // The read is successful; update the counters.
      g_nReadsInProgress++;
      g_nMaxReadsInProgress = 
         max(g_nMaxReadsInProgress, g_nReadsInProgress);
   } else {
      // The read is unsuccessful.
      // Get the reason for the failure.
      dwLastError = GetLastError();

      if (ERROR_HANDLE_EOF == dwLastError) {
         // Read failed because of an attempt to read past the
         // end of the file; set global status indicator.
         g_CopyStatus = csDoneReading;
         AddStr(__TEXT("%d: Read past end-of-file."),
            (int) lpOverlapped->hEvent);
      } else {
         // Read failed because of another error; 
         // set global status indicator and global error.
         g_CopyStatus = csError;
         g_dwLastError = dwLastError;

         AddStr(__TEXT("%d: Read caused an error (%d)."),
            (int) lpOverlapped->hEvent, g_dwLastError);
      }
   }
}


/////////////////////////////////////////////////////////////


VOID WINAPI InputCompletion (DWORD fdwError,
   DWORD cbTransferred, LPOVERLAPPED lpOverlapped) {

   BOOL fWriteOk;

   // Signal that a read has been completed.
   g_nReadsInProgress--;

   switch (fdwError) {
      case 0:
         // The read has been completed successfully.
         g_nCompletedReads++;

         AddStr(__TEXT("%d: Write, Offset=%d, Len=%d."),
            (int) lpOverlapped->hEvent, lpOverlapped->Offset, 
            cbTransferred);
      
         // Write this buffer to the destination file.
         // The OVERLAPPED structure contains the offset
         // the buffer was read from. This is the same offset
         // the buffer should be written to.
         fWriteOk = WriteFileEx(g_hFileDst,
            g_bBuffers + (int) lpOverlapped->hEvent * BUFFSIZE,
            cbTransferred, lpOverlapped, OutputCompletion);

         if (fWriteOk) {
            // The write is successful; update the counters.
            g_nWritesInProgress++;
            g_nMaxWritesInProgress = max(
               g_nMaxWritesInProgress, g_nWritesInProgress);
         } else {
            // The write is unsuccessful.
            // Get the reason for the failure, and set global 
            // status indicator and global error.
            g_CopyStatus = csError;
            g_dwLastError = GetLastError();

            AddStr(__TEXT("%d: Write caused an error (%d)."),
               (int) lpOverlapped->hEvent, g_dwLastError);
         }
         break;
         
      case ERROR_HANDLE_EOF:  
         // Read past the end of the file.
         // Update the global status indicator.
         g_CopyStatus = csDoneReading;
         AddStr(__TEXT("%d: Done reading source file."),
            (int) lpOverlapped->hEvent);
         break;
   }
}


/////////////////////////////////////////////////////////////


VOID WINAPI OutputCompletion (DWORD fdwError, 
   DWORD cbTransferred, LPOVERLAPPED lpOverlapped) {

   // Signal that a write has been completed.
   g_nWritesInProgress--;

   if (fdwError == 0) {
      // The write has been completed successfully.
      g_nCompletedWrites++;

      // Start reading the next chunk of the file if we're not
      // done reading the file yet.
      if (csCopying == g_CopyStatus) {
         ReadNext(lpOverlapped);
      }

      // Test to see whether this was the last write, and if
      // so, update the global status indicator so that the
      // main copy loop will know the file has been copied.
      if ((g_CopyStatus == csDoneReading) && 
          (g_nWritesInProgress == 0)) {
         g_CopyStatus = csDoneWriting;
         AddStr(__TEXT("%d: Done writing destination file."),
            (int) lpOverlapped->hEvent);
      }
   }
}


/////////////////////////////////////////////////////////////


BOOL FileCopy (LPCTSTR pszFileSrc, LPCTSTR pszFileDst) {
   int nBuffer;

   // Open the existing source file for input.
   g_hFileSrc = CreateFile(pszFileSrc, GENERIC_READ,
      FILE_SHARE_READ, NULL, OPEN_EXISTING, 
      FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED | 
      FILE_FLAG_SEQUENTIAL_SCAN, NULL);

   if (g_hFileSrc == INVALID_HANDLE_VALUE) {
      g_dwLastError = GetLastError();
      return(FALSE);
   }
   
   // Create the new destination file for output.
   g_hFileDst = CreateFile(pszFileDst, GENERIC_WRITE,
      0, NULL, CREATE_ALWAYS, 
      FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED | 
      FILE_FLAG_SEQUENTIAL_SCAN, NULL);

   if (g_hFileDst == INVALID_HANDLE_VALUE) {
      CloseHandle(g_hFileSrc);
      g_dwLastError = GetLastError();
      return(FALSE);
   }

   // Get ready to start copying the file.
   g_CopyStatus = csCopying;
   g_dwNextReadOffset     = 0;
   g_nReadsInProgress     = 0;
   g_nMaxReadsInProgress  = 0;
   g_nCompletedReads      = 0;
   g_nWritesInProgress    = 0;
   g_nMaxWritesInProgress = 0;
   g_nCompletedWrites     = 0;

   // Start the copy engine by allowing the buffers to begin 
   // reading data from the file.
   for (nBuffer = 0; nBuffer < BUFFNUM; nBuffer++) {
      g_Overlapped[nBuffer].hEvent = (HANDLE) nBuffer;
      ReadNext(&g_Overlapped[nBuffer]);
   }


   // Loop until an error has occurred or until the
   // destination file has been written.
   while ((g_CopyStatus != csError) &&
          (g_CopyStatus != csDoneWriting)) {
        
      // Put this thread to sleep until it's awakened by 
      // an alertable file I/O completion.
      SleepEx(INFINITE, TRUE);
   }

   CloseHandle(g_hFileDst);
   CloseHandle(g_hFileSrc);
   
   if (g_CopyStatus == csError) {
      SetLastError(g_dwLastError);
      AddStr(__TEXT("File copy error %d."), g_dwLastError);      
   } else {
      AddStr(__TEXT("File copied successfully."));
   }

   // Put some statistical information in the list box.
   AddStr(__TEXT("Max reads in progress=%d."), 
      g_nMaxReadsInProgress);
   AddStr(__TEXT("Completed reads=%d."), g_nCompletedReads);
   AddStr(__TEXT("Max writes in progress=%d."), 
      g_nMaxWritesInProgress);
   AddStr(__TEXT("Completed writes=%d."), g_nCompletedWrites);
      
   return(g_CopyStatus != csError);
}


/////////////////////////////////////////////////////////////


BOOL Dlg_OnInitDialog (HWND hwnd, HWND hwndFocus, 
   LPARAM lParam) {

   // Save the handle of the dialog's list box in a global 
   // so that the threads can easily gain access to it.
   g_hwndLB = GetDlgItem(hwnd, IDC_LOG);


   // Associate an icon with the dialog box.
   SetClassLong(hwnd, GCL_HICON, (LONG)
      LoadIcon((HINSTANCE) GetWindowLong(hwnd, GWL_HINSTANCE),
      __TEXT("AlertIO")));


   // Disable the "Copy" button because no file 
   // has been selected yet.
   EnableWindow(GetDlgItem(hwnd, IDOK), FALSE);
   
   return(TRUE);
}


/////////////////////////////////////////////////////////////


void Dlg_OnCommand (HWND hwnd, int id, 
   HWND hwndCtl, UINT codeNotify) {
   
   TCHAR szPathname[_MAX_DIR];
   BOOL fOk;
   OPENFILENAME ofn;

   switch (id) {
      case IDOK:
         // Copy the source file to the destination file.
         ListBox_ResetContent(g_hwndLB);
         Static_GetText(GetDlgItem(hwnd, IDC_SRCFILE), 
            szPathname, sizeof(szPathname));
         SetCursor(LoadCursor(NULL, IDC_WAIT));
         FileCopy(szPathname, DSTFILENAME);
         break;

      case IDC_BROWSE:
         memset(&ofn, 0, sizeof(ofn));
         ofn.lStructSize = sizeof(ofn);
         ofn.hwndOwner = hwnd;
         ofn.lpstrFilter = __TEXT("*.*\0");

         _tcscpy(szPathname, __TEXT("*.*"));
         ofn.lpstrFile = szPathname;

         ofn.nMaxFile = sizeof(szPathname);
         ofn.Flags = OFN_FILEMUSTEXIST;
         fOk = GetOpenFileName(&ofn);
         
         if (fOk) {
            HANDLE hFile;
            Static_SetText(GetDlgItem(hwnd, IDC_SRCFILE), 
               szPathname);
            hFile = CreateFile(szPathname, GENERIC_READ,
               0, NULL, OPEN_EXISTING, 0, NULL);

            SetDlgItemInt(hwnd, IDC_SRCFILESIZE, 
               GetFileSize(hFile, NULL), FALSE);
            CloseHandle(hFile);
         }
         
         // Enable the "Copy" button if the user selected
         // a valid pathname.
         GetWindowText(GetDlgItem(hwnd, IDC_SRCFILE), 
            szPathname, sizeof(szPathname));
         EnableWindow(GetDlgItem(hwnd, IDOK), 
            szPathname[0] != __TEXT('('));
          
         if (fOk) {
            // If the user pressed the OK button in the file
            // dialog box, change focus to the "Copy" button.
            FORWARD_WM_NEXTDLGCTL(hwnd, GetDlgItem(hwnd, IDOK), TRUE, SendMessage);
          }
         break;
         
      case IDCANCEL:
         EndDialog(hwnd, id);
         break;
   }
}

/////////////////////////////////////////////////////////////


BOOL CALLBACK Dlg_Proc (HWND hDlg, UINT uMsg, 
   WPARAM wParam, LPARAM lParam) {

   BOOL fProcessed = TRUE;

   switch (uMsg) {
      HANDLE_MSG(hDlg, WM_INITDIALOG, Dlg_OnInitDialog);
      HANDLE_MSG(hDlg, WM_COMMAND, Dlg_OnCommand);

      default:
         fProcessed = FALSE;
         break;
   }

   return(fProcessed);
}


/////////////////////////////////////////////////////////////


int WINAPI WinMain (HINSTANCE hinstExe,
   HINSTANCE hinstPrev, LPSTR lpszCmdLine, int nCmdShow) {
   
   DialogBox(hinstExe, MAKEINTRESOURCE(IDD_ALERTIO), 
      NULL, Dlg_Proc);

   return(0);
}


//////////////////////// End Of File ////////////////////////
