/*************************************************************
Module name: FileRev.C
Notices: Copyright (c) 1995 Jeffrey Richter
*************************************************************/


#include "..\AdvWin32.H"     /* See Appendix B for details. */
#include <windows.h>
#include <windowsx.h>

#pragma warning(disable: 4001)       /* Single-line comment */

#include <tchar.h>
#include <string.h>            // For _strrev
#include "Resource.H"


//////////////////////////////////////////////////////////////


#define FILENAME  __TEXT("FILEREV.DAT")


//////////////////////////////////////////////////////////////


int WINAPI WinMain (HINSTANCE hinstExe,
   HINSTANCE hinstPrev, LPSTR lpszCmdLine, int nCmdShow) {

   HANDLE hFile, hFileMap;
   LPVOID lpvFile;
   LPSTR lpchANSI;      // Always ANSI
   LPWSTR lpchUnicode;  // Always Unicode 
   BOOL fIsTextUnicode = FALSE;
   DWORD dwFileSize;
   LPTSTR lpszCmdLineT;

   STARTUPINFO si = { 0 };
   PROCESS_INFORMATION pi;

   // Get the name of the file the user wants to reverse.
   // We must use GetCommandLine here instead of WinMain's 
   // lpszCmdLine parameter because lpszCmdLine is always an 
   // ANSI string, never a Unicode string. GetCommandLine 
   // returns ANSI or Unicode depending 
   // on how we've compiled.
   lpszCmdLineT = _tcschr(GetCommandLine(), __TEXT(' '));

   if (lpszCmdLineT != NULL) {
      // We found a space after the executable file's name.  
      // Now let's skip over any white space to get to 
      // the first argument.
      while (*lpszCmdLineT == __TEXT(' '))
         lpszCmdLineT++;
   }

   if ((lpszCmdLineT == NULL) || (*lpszCmdLineT == 0)) {
      // If no space was found or there are no arguments
      // after the executable file's name, display an
      // error message.
      MessageBox(NULL, 
         __TEXT("You must enter a filename on ")
         __TEXT("the command line."),
         __TEXT("FileRev"), MB_OK);
      return(0);
   }

   // Copy input file to FILEREV.DAT so that the original is 
   // not destroyed. We must use GetCommandLine here instead 
   // of lpszCmdLine because the lpszCmdLine passed to WinMain
   // is always ANSI, never Unicode.
   if (!CopyFile(lpszCmdLineT, FILENAME, FALSE)) {
      // Copy failed.
      MessageBox(NULL, 
         __TEXT("New file could not be created."),
         __TEXT("FileRev"), MB_OK);
      return(0);
   }

   // Open the file for reading and writing.
   hFile = CreateFile(FILENAME, GENERIC_WRITE | GENERIC_READ,
      0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

   if (hFile == INVALID_HANDLE_VALUE) {
      // File open failed.
      MessageBox(NULL, __TEXT("File could not be opened."),
         __TEXT("FileRev"), MB_OK);
      return(0);
   }

   // Get the size of the file. I am assuming here that the
   // file is smaller than 4 GB.
   dwFileSize = GetFileSize(hFile, NULL);

   // Create the file-mapping object. The file-mapping object 
   // is 1 character bigger than the file size so that a zero
   // character can be placed at the end of the file to terminate
   // the string (file). Because I don't yet know if the file
   // contains ANSI or Unicode characters, I assume worst case
   // and add the size of a WCHAR instead of CHAR.
   hFileMap = CreateFileMapping(hFile, NULL, PAGE_READWRITE,
      0, dwFileSize + sizeof(WCHAR), NULL);

   if (hFileMap == NULL) {
      // File-Mapping open failed.
      MessageBox(NULL, __TEXT("File map could not be opened."),
         __TEXT("FileRev"), MB_OK);
      CloseHandle(hFile);
      return(0);
   }

   // Get the address where the first byte of the file 
   // is mapped into memory.
   lpvFile = MapViewOfFile(hFileMap, FILE_MAP_WRITE, 0, 0, 0);

   if (lpvFile == NULL) {
      // Map view of file failed.
      MessageBox(NULL, __TEXT("Could not map view of file."),
         __TEXT("FileRev"), MB_OK);
      CloseHandle(hFileMap);
      CloseHandle(hFile);
      return(0);
   }


   // If we are not running on Windows NT 3.10, check to see
   // whether the text file is Unicode otherwise assume ANSI.
   if (LOWORD(GetVersion()) != 0x0A03) {
      // Take an educated guess as to whether the text file
      // contains ANSI or Unicode characters.
      fIsTextUnicode = IsTextUnicode(lpvFile, dwFileSize,
         NULL);
   }

   if (!fIsTextUnicode) {
      // For all the file manipulations below, we explicitly 
      // use ANSI functions instead of Unicode functions 
      // because, although the application can be ANSI or 
      // Unicode, it is processing an ANSI file.

      // Put a zero character at the very end of the file.
      lpchANSI = (LPSTR) lpvFile;
      lpchANSI[dwFileSize] = 0;

      // Reverse the contents of the file.
      _strrev(lpchANSI);

      // Convert all "\n\r" combinations back to "\r\n" to 
      // preserve the normal end-of-line sequence.
      lpchANSI = strchr(lpchANSI, '\n'); // Find first '\n'.

      while (lpchANSI != NULL) {
         // We have found an occurrence....
         *lpchANSI++ = '\r';   // Change '\n' to '\r'.
         *lpchANSI++ = '\n';   // Change '\r' to '\n'.
         lpchANSI = strchr(lpchANSI, '\n'); // Find the next
                                            // occurrence.
      }

   } else {
      // For all the file manipulations below, we explicitly 
      // use Unicode functions instead of ANSI functions 
      // because, although the application can be ANSI or 
      // Unicode, it is processing a Unicode file.

      // Put a zero character at the very end of the file.
      lpchUnicode = (LPWSTR) lpvFile;
      lpchUnicode[dwFileSize] = 0;

      // Reverse the contents of the file.
      _wcsrev(lpchUnicode);

      // Convert all "\n\r" combinations back to "\r\n" to 
      // preserve the normal end-of-line sequence.
      lpchUnicode = wcschr(lpchUnicode, L'\n'); // Find first
                                                // '\n'.

      while (lpchUnicode != NULL) {
         // We have found an occurrence....
         *lpchUnicode++ = L'\r';   // Change '\n' to '\r'.
         *lpchUnicode++ = L'\n';   // Change '\r' to '\n'.
         lpchUnicode = wcschr(lpchUnicode, L'\n'); // Find the
                                                   // next
                                                   // occurrence.
      }
   }


   // Clean up everything before exiting.
   UnmapViewOfFile(lpvFile);
   CloseHandle(hFileMap);

   // Remove the trailing zero byte added earlier by 
   // positioning the file pointer at the end of the file,
   // not including the zero byte and, setting
   // the end-of-file.
   SetFilePointer(hFile, dwFileSize, NULL, FILE_BEGIN);

   // SetEndOfFile must be called after the
   // file-mapping kernel object is closed.
   SetEndOfFile(hFile);
   CloseHandle(hFile);

   // Spawn Notepad to see the fruits of our labors.
   si.cb = sizeof(si);
   si.wShowWindow = SW_SHOW;
   si.dwFlags = STARTF_USESHOWWINDOW;
   if (CreateProcess(NULL, _TEXT("NOTEPAD.EXE ") FILENAME,
      NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi)) {

      CloseHandle(pi.hThread);
      CloseHandle(pi.hProcess);
   }

   return(0);
}


/////////////////////// End Of File ///////////////////////
